/*
     This file is part of GNUnet.
     (C) 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/testbed/gnunet-testbed.c 
 * @brief A Testbed tool for performing distributed experiments
 * @author Ronaldo Alves Ferreira
 * @author Christian Grothoff
 * @author Murali Krishan Ramanathan
 *
 * Todo:
 * - implement more messages
 * - extend scriptability (add labels and loops; requires *recording* of the commands!)
 * - test
 **/

#include "testbed.h"

#define TESTBED_VERSION		"0.0.1"
#define MAX_CMD_LINE		250

/**
 * @brief struct keeping per-peer information for the testbed
 */
typedef struct {
  /** IP address of the peer */
  IPaddr	     ip;
  /** CS port of the peer */
  unsigned short     port;
  /** string describing the peer address */
  char	* ips;
  /** socket to communicate with the peer */
  GNUNET_TCP_SOCKET sock;
  /** HELO message identifying the peer in the network */
  HELO_Message	    *helo;
} NODE_INFO;

/**
 * List of nodes known to the testbed.
 */
static NODE_INFO * nodes = NULL;

/*
 * Number of known nodes, size of the nodes array.
 */
static int nnodes = 0;

/*
 * Number of active connections
 */
static int connections;

/**
 * What is the plan for this one? -- CG
 */
static char * confilename;

/**
 * Should the driver exit?
 */
static int do_quit = NO;

/**
 * Shell parser and interpreter.  Runs the command
 * given.
 */
static void run(char *cmdline);

/* ********** implementation of commands ********** */

/**
 * Add a node to the configuration.
 */
static void addNode(char * line) {
  int currindex;
  TESTBED_ADD_PEER_MESSAGE * hdr;
  char ip_aux[30];
  TESTBED_CS_MESSAGE *msg;
  char * s = line;
    
  printf("Adding node: %s\n", s);
    
  /* FIXME: the code below needs to be improved. */
  while (*s != ':') s++;
  *s++ = '\0';
  strcpy(ip_aux, line);
  
  currindex = nnodes;
  GROW(nodes,
       nnodes,
       nnodes+1);
  nodes[currindex].ips = STRDUP(ip_aux);
  nodes[currindex].port = atoi(s);
  inet_aton(ip_aux, &nodes[currindex].ip.addr);
  printf("IP: %s Port:%d\n", 
	 ip_aux, 
	 nodes[currindex].port);
  
  if (SYSERR != initGNUnetClientSocket(nodes[currindex].port,
				       ip_aux, 
				       &nodes[currindex].sock)) {     
    TESTBED_GET_HELO_MESSAGE *phelo;
    
    msg = MALLOC(sizeof(TESTBED_GET_HELO_MESSAGE));
    msg->header.size = htons(sizeof(TESTBED_GET_HELO_MESSAGE));
    msg->header.tcpType = htons(TESTBED_CS_PROTO_REQUEST);
    msg->msgType = htonl(TESTBED_GET_HELO);
    phelo = (TESTBED_GET_HELO_MESSAGE *)msg->data;
    phelo->proto = htons(TCP_PROTOCOL_NUMBER);
    if (SYSERR == writeToSocket(&nodes[currindex].sock, 
				&msg->header)) {
      printf("Could not send message to node %s.  Trying other available peers.\n", 
	     nodes[currindex].ips);
      FREE(msg);
      FREE(nodes[currindex].ips);
      GROW(nodes, 
	   nnodes,
	   nnodes-1);
      return;
    }
    FREE(msg);
    
    printf("writeToSocket returned successfully\n");
    hdr = NULL;      
    if (OK == readFromSocket(&nodes[currindex].sock, 
			     (CS_HEADER**)&hdr)) {
      nodes[currindex].helo = MALLOC(HELO_Message_size(&hdr->helo));
      memcpy(nodes[currindex].helo, 
	     &hdr->helo, 
	     HELO_Message_size(&hdr->helo));
      printf("HELO message received\n"\
	     "\tprotocol: %d\n"\
	     "\tMTU: %u\n",
	     ntohs(hdr->helo.protocol), 
	     (unsigned int) ntohl(hdr->helo.MTU));
      FREE(hdr);
    } else {
      LOG(LOG_WARNING,
	  "WARNING: failed to receive HELO from peer!\n");
      FREE(nodes[currindex].ips);
      GROW(nodes, 
	   nnodes,
	   nnodes-1);
      return;
    }
  }
}


/* */
static int sendMessage(unsigned int msgType, 
		       int dst, 
		       unsigned int argSize,
		       void * arg) {
  TESTBED_CS_MESSAGE * msg;
  int msgsz;

  if (dst >= nnodes || dst < 0) {
    printf("Invalid destination node %d\n", dst);
    return SYSERR;
  }
  /* Nodes are valid. Send message to delete connection. */      
  msgsz = sizeof(TESTBED_CS_MESSAGE)+argSize;
  msg = MALLOC(msgsz);
  msg->header.size = htons(msgsz);
  msg->header.tcpType = htons(TESTBED_CS_PROTO_REQUEST);
  msg->msgType = htonl(msgType);
  memcpy(&msg->data[0],
	 arg,
	 argSize);
  printf("Sending message to node %s\n", 
	 nodes[dst].ips);
  if (SYSERR == writeToSocket(&nodes[dst].sock, 
			      &msg->header))
    printf("Could not send message to node %s\n",
	   nodes[dst].ips);
  else
    printf("writeToSocket returned successfully\n"); 
  FREE(msg);
  return OK;
}

/* */
static void delConnection(char *cmdline) {
  int src, dst;
  char *s = cmdline;

  sscanf (s, "%d %d", &src, &dst);
  printf("del %d->%d\n", src, dst);
  if (OK == sendMessage(TESTBED_DEL_PEER, 
			src,
			sizeof(HostIdentity),
			&nodes[dst].helo->senderIdentity)) {
    /* FIXME: check return value from peer! */
    connections--;
  }
}

/* */
static void addConnection(char *cmdline) {
  int src, dst;
  char *s = cmdline;
  
  sscanf (s, "%d %d", &src, &dst);
  printf("add %d->%d\n", src, dst);
  if (OK == sendMessage(TESTBED_ADD_PEER,
			src,
			dst,
			0)) {
    /* FIXME: check a return value from
       the peer! */
    connections++;
  }
}

/* */
static void setTrust(char *cmdline) {
  int src, dst, value;
  int confirmation;
  char *s = cmdline;
  
  sscanf (s, "%d %d %d", &src, &dst, &value);
  printf("set-trust %d->%d %d\n", src, dst, value);
  // sendMessage(TESTBED_SET_TVALUE, dst, value, SIZE, MSG);

  if (OK == readTCPResult(&nodes[src].sock,
			  &confirmation)) {
    if (confirmation == OK) {
      printf("Trust value between %d and %d set to %d\n", 
	     src, 
	     dst, 
	     value);
    } else {
      LOG(LOG_WARNING,
	  "WARNING: peer encountered error setting trust value.\n");
    }
  } else {
    LOG(LOG_ERROR,
	"ERROR: peer did not confirm setting trust value.\n");
  }  
}

/**
 * ATTN RF: I've hacked this method to show a bit more how the
 * protocol should work IMO.  You may want to model some of the others
 * accordingly. -- CG
 */
static void getTrust(char *cmdline) {
  int   src, dst, value;
  char *s = cmdline;
  
  sscanf (s, "%d %d", &src, &dst);
  if (OK == sendMessage(TESTBED_GET_TVALUE, 
			src, 
			sizeof(HostIdentity),
			&nodes[dst].helo->senderIdentity)) {
    if (OK == readTCPResult(&nodes[src].sock,
			    &value)) {
      setConfigurationInt("GNUNET-TESTBED-VARIABLE",
			  "$0",
			  value);
      LOG(LOG_INFO,
	  "INFO: Trust value between %d and %d is %d\n", 
	  src, 
	  dst, 
	  value);
    } else {
      LOG(LOG_ERROR,
	  "ERROR: could not get trust value.\n");
      setConfigurationInt("GNUNET-TESTBED-VARIABLE",
			  "$0",
			  -1);
    }
  } else {
    LOG(LOG_ERROR,
	"ERROR: could not send request for trust value.\n");
  }
}

/**
 * Helper function for (un)loadModule.
 * @param type load or unload requested?
 */
static void XXloadModule(unsigned short type,
			 char * cmdLine) {
  int ok;
  int dst;
  char * modulename;

  modulename = cmdLine;
  while ( ( (*modulename) != '\0') &&
	  ( (*modulename) != ' ') )
    modulename++;
  if ( *modulename != ' ') {
    printf("Input error.  Format: load-module PEERID MODULENAME.\n");
    return;
  }
  *modulename = '\0';
  modulename++;
  if (strlen(modulename) == 0) {
    printf("Input error.  You must specify a module name.\n");
    return;
  }
  if (1 != sscanf(cmdLine, "%d", &dst)) {
    printf("Input error.  PEERID must be number.\n");
    return;
  }   
  ok = sendMessage(type,
		   dst,
		   strlen(modulename),
		   modulename);
  if (ok == OK) {
    if (OK == readTCPResult(&nodes[dst].sock,
			    &ok)) {
      if (ok != OK)
	LOG(LOG_WARNING,
	    "WARNING: peer had error loading module.\n");
    } else {
      LOG(LOG_ERROR,
	  "ERROR: peer did not send confirmation message for loadModule.\n");
    }
  } else {
    LOG(LOG_ERROR,
	"ERROR: could not send loadModule message.\n");
  }
}

static void loadModule(char * cmdLine) {
  XXloadModule(TESTBED_LOAD_MODULE,
	       cmdLine);
}
static void unloadModule(char * cmdLine) {
  XXloadModule(TESTBED_UNLOAD_MODULE,
	       cmdLine);
}

/* */
static void setBW(char *cmdline)
{
    int src, dst, in, out;
    char *s = cmdline;
    
    sscanf(s,
	   "%d %d %d %d",
	   &src, 
	   &dst, 
	   &in, 
	   &out);
    printf("set-trust %d->%d in=%d out=%d\n", 
	   src, 
	   dst,
	   in, 
	   out);
    /*    if (OK == sendMessage(TESTBED_SET_IN_BW, 
			  src,
			  dst, 
			  in))
      sendMessage(TESTBED_SET_OUT_BW, 
		  src, 
		  dst, 
		  out);
    */
}

/* */
static void setLoss(char *cmdline) {
  int src, dst, value;
  char *s = cmdline;
  
  sscanf (s, "%d %d %d",
	  &src, 
	  &dst,
	  &value);
  printf("set-loss %d->%d %d\n", 
	 src, 
	 dst, 
	 value);
  /* sendMessage(TESTBED_SET_LOSS_RATE,
     src,
     dst,
     value);*/
}

/* */
static void getStats(char *cmdline) {
  int   src, dst;
  char *s = cmdline;
  
  sscanf (s, "%d %d", &src, &dst);
  /*
  printf("get-trust %d->%d\n", src, dst);  
  if (OK == sendMessage(TESTBED_GET_TVALUE, src, dst, 0)) {
    CS_HEADER *hdr;
    hdr = MALLOC(5000); 
    
    readFromSocket(nodes[src].sock, (CS_HEADER **)&hdr);
    intptr = (int *)hdr->data;
    value = htonl(*intptr);
    printf("Trust value between %d and %d: %d\n", src, dst, value);
    FREE(hdr);
  }
*/
}


static void printOnlineHelp(char * args) {
  printf("nodes filename          ; load nodes in the network from file\n"\
	 "connections filename    ; load the connections among the nodes from file\n"\
	 "delete filename         ; load delete information from file\n"\
	 "showups                 ; show the nodes that are up\n"\
	 "add src dst             ; add a connection from node src to dst\n"\
	 "del src dst             ; delete a connection from node src to dst\n"\
	 "set-trust src dst value ; set value as the trust value in src for dst\n"\
	 "get-trust src dst       ; get trust value in src for dst\n"\
	 "set-bw src dst in out   ; set in and out bandwidth between src and dst\n"
	 "set-loss src dst loss   ; set the loss probability between src and dst\n"
	 "get-stats peer          ; get peer statistics\n"
	 );
}

static void showUps(char * args) {
  int i;
  for (i = 0; i < nnodes; i++)
    printf("%d - %s:%d\n", 
	   i+1, 
	   nodes[i].ips, 
	   nodes[i].port);
}

static void doExit(char * args) {
  do_quit = YES;
}

/**
 * Simple arithmetic evaluation.  Uninitialized variables are 0.
 * The trivial (prefix) grammar:
 *
 * OP = +|-|*|/|%
 * EXPR = OP EXPR EXPR|VARIABLE|INTEGER
 *
 * @param args an EXPRession to evaluate.
 */
static int eval(char * args) {
  char * eptr;
  char * snd;
  long int val;

  while (*args == ' ')
    args++; /* skip spaces */

  if (*args == '\0') {
    printf("Empty string passed as expression.  Using 0.\n");
    return 0;
  }
  
  val = strtol(args, &eptr, 0);  
  if (*eptr == '\0')
    return (int) val;
  eptr = args+1; /* skip prefix operator */
  snd = eptr;
  while ( (*snd != '\0') &&
	  (*snd != ' ') ) snd++;
  if ( (snd == eptr) || (*snd == '\0') ) 
    return getConfigurationInt("GNUNET-TESTBED-VARIABLE",
			       args);  
  *snd = '\0';
  snd++;

  switch (args[0]) {
  case '+': 
    return eval(eptr) + eval(snd);
  case '-': 
    return eval(eptr) - eval(snd);
  case '*': 
    return eval(eptr) * eval(snd);
  case '/': 
    return eval(eptr) / eval(snd);
  case '%': 
    return eval(eptr) % eval(snd);
  default:
    return getConfigurationInt("GNUNET-TESTBED-VARIABLE",
			       args);
  }    
}

/**
 * "if X CMD" runs "CMD" if "X" evaluates to non-zero.  X must be a
 * simple variable or a numeric primitive (may not contain spaces).
 */
static void pred(char * args) {
  char * cond;
  cond = args;
  while ( (*args != '\0') &&
	  (*args != ' ') ) args++;
  if (*args == '\0') {
    printf("You must specify a value.\n");
    return;
  }
  *args = '\0';
  args++;
  if (eval(cond))
    run(args);
}

/**
 * "let A EXPR" sets the value of variable A to whatever EXPR evaluates
 * to.
 */
static void let(char * args) {
  char * variable;
  variable = args;
  while ( (*args != '\0') &&
	  (*args != ' ') ) args++;
  if (*args == '\0') {
    printf("You must specify a value.\n");
    return;
  }
  *args = '\0';
  args++;
  
  setConfigurationInt("GNUNET-TESTBED-VARIABLE",
		      variable,
		      eval(args));  
}

/* ********** script processing ************** */

typedef void (*CommandHandler)(char * arguments);

typedef struct CMD_ENTRY_ {
  char * command;
  CommandHandler handler;
} CMD_ENTRY;

static CMD_ENTRY commands[];

/**
 * Run the given command handler on every line
 * of the given file.
 */
static void sourceScript(char * filename,
			 CommandHandler r) {
  char line[MAX_CMD_LINE];
  FILE * config = fopen(filename, "r");

  if (!config) {
    printf ("Could not open file: %s\n", 
	    filename);
    return;
  }
  while (fgets(line, MAX_CMD_LINE, config) > 0) {
    char * s = line;
    s = line;
    while (*s == ' ') s++;
    r(s);
  }
  fclose(config);
} 

/**
 * The first argument in cmdline must be ANOTHER
 * command (like, say "add-node").  That command
 * is then executed for EACH line that is found
 * in the file specified by the second argument.
 *
 * For example, "source add-node myPeerList"
 * will run the command "add-node" on every
 * IP:PORT line in the file myPeerList.
 *
 * Another good example is "source run myscript",
 * which will run the commands specified in
 * myscript.
 */
static void source(char * cmdline) {
  int i=0;
  while (*cmdline == ' ') cmdline++;
  while (commands[i].command != NULL) {
    int sl = strlen(commands[i].command);
    if ( (0 == strncmp(cmdline, 
		       commands[i].command,
		       sl)) &&
	 (cmdline[sl] == ' ') ) {
      while (cmdline[sl] == ' ')
	sl++;
      sourceScript(&cmdline[sl],
		   commands[i].handler);
      return;
    }
    i++;
  }
  printf("Unknown command %s\n", 
	 cmdline);
}


/* *********** main interpreter loop ************ */

static CMD_ENTRY commands[] = {
  { "help", &printOnlineHelp },
  { "get-trust", &getTrust },
  { "set-bw", &setBW },
  { "set-trust", &setTrust },
  { "add-node", &addNode },  
  { "add", &addConnection },
  { "del", &delConnection },
  { "exit", &doExit },
  { "showups", &showUps}, 
  { "set-loss", &setLoss} ,
  { "get-stats", &getStats}, 
  { "source", &source },
  { "load-module", &loadModule },
  { "unload-module", &unloadModule },
  { "run", &run },
  { "let", &let },
  { "if", &pred },
  { NULL, NULL }, /* termination */
};


/**
 * Shell parser and interpreter.  Runs the command
 * given.
 */
static void run(char *cmdline) {
  int i = 0;

  while (*cmdline == ' ') cmdline++;
  while (commands[i].command != NULL) {
    int sl = strlen(commands[i].command);
    if ( (0 == strncmp(cmdline, 
		       commands[i].command,
		       sl)) &&
	 (cmdline[sl] == ' ') ) {
      while (cmdline[sl] == ' ')
	sl++;
      commands[i].handler(&cmdline[sl]);
      return;
    }
    i++;
  }
  printf("Unknown command\n");
}

/* *************** command line options *********** */

/* */
static void testbedUsage() {
  printf("-t FILENAME, --tbed=FILENAME  : "
	 "file with all configuration parameters for the Testbed\n");
}

/**
 * Parse the options, set the timeout.
 *
 * @param argc the number of options
 * @param argv the option list (including keywords)
 * @return OK on error, SYSERR if we should exit 
 **/
static int parseOptions(int argc,
			char ** argv) {
  int option_index;
  int c;  
  
  FREENONNULL(setConfigurationString("GNUNETD",
				     "LOGFILE",
				     NULL));
  while (1) {
    static struct GNoption long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { "tbed", 0, 0, 'c' },
      { 0,0,0,0 }
    };    
   
    option_index = 0;
    c = GNgetopt_long(argc,
		      argv, 
		      "vhdc:L:t:", 
		      long_options, 
		      &option_index);    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, GNoptarg))
      continue;
    switch(c) {
    case 'v': 
      printf("GNUnet v%s, gnunet-testbed v%s\n",
	     VERSION,
	     TESTBED_VERSION);
      return SYSERR;

    case 'h': 
      printf("USAGE: gnunet-testbed\n");
      printHelpDefault();
      testbedUsage();
      return SYSERR;
    case 't':
      confilename = STRDUP(GNoptarg);
      break;      
    default: 
      LOG(LOG_FAILURE,
	  "FAILURE: Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      return -1;
    } /* end of parsing commandline */
  } /* while (1) */
  return OK;
}

/* **************** main **************** */

/**
 * Create a testbed environment with multiple peers to collect 
 * GNUnet statistics.
 *
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return return value from gnunetsearch: 0: ok, -1: error
 **/   
int main(int argc, char ** argv) {
  char cmdline[MAX_CMD_LINE];

  if (SYSERR == initUtil(argc, argv, &parseOptions))
    return 0; 
  connections = 0;
  while (do_quit == NO) {
    int slen;
    
    printf("> ");
    fgets (cmdline, MAX_CMD_LINE, stdin);
    slen = strlen(cmdline);
    if (cmdline[slen-1] == '\n')
      cmdline[slen-1] = '\0';
    printf("Command line: \"%s\"\n\n", cmdline);
    run(cmdline);
  }
  printf("Exiting.\n");
  doneUtil();
  return 0;
}

/* end of gnunet-testbed.c */ 
